/* Written by Yong Li, Depeng Liu                                       */
/* Copyright (c) 2016                  	                               */
/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
 /* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License    */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.*/

package cn.ac.ios.learner.multiplicity;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;



import cn.ac.ios.automata.Automaton;
import cn.ac.ios.automata.MultiplicityAutomaton;
import cn.ac.ios.automata.words.ContextWord;
import cn.ac.ios.automata.words.Word;
import cn.ac.ios.learner.multiplicity.InputHelper;
import cn.ac.ios.learner.Learner;
import cn.ac.ios.learner.LearnerType;
import cn.ac.ios.table.ObservationRow;
import cn.ac.ios.query.MembershipOracle;
import cn.ac.ios.query.Query;
import cn.ac.ios.query.QuerySimple;

import Jama.Matrix;

public class LearnerMultiplicity implements Learner<Automaton, Double> {

	private final ContextWord contextWord;
	private ObservationTableMult observationTable;
	private final MembershipOracle<Double> membershipOracle;
	private boolean alreadyStarted = false;
	//consisting of Fx for every x in upper table
	private Matrix A;
	//matrixes:list of matrixes consisting of Fxa for every x in upper table and a specific 'a' in alphabet
	//mat: list of matrixes of u(a)
	private List<Matrix> matrixes, mat; 
	// size of upper table
	private int size;
	// index of column in the counterexample
	private int colIndex;
	// reserving the weights of states
	private Matrix gama;
	
	public LearnerMultiplicity(ContextWord contextWord, MembershipOracle<Double> membershipOracle){
		assert contextWord != null && membershipOracle != null;
		this.contextWord = contextWord;
		this.membershipOracle = membershipOracle;
		this.observationTable = new ObservationTableMult(contextWord);
		// add every alphabet
		for(int letterNr = 0; letterNr < this.contextWord.getAlphabet().size(); letterNr ++) {
//			observationTable.addLowerRow(contextWord.getLetterWord(letterNr));
		}
		size = observationTable.getsize();
	}
	
	
	@Override
	public LearnerType getLearnerType() {
		return LearnerType.WEIGHT;
	}
	@Override
	public void startLearning() {
		if(alreadyStarted )
			try {
				throw new Exception("Learner should not be started twice");
			} catch (Exception e) {
				e.printStackTrace();
			}
		alreadyStarted = true;
		// ask initial queries for upper table
//		processMembershipQueries(observationTable.getUpperTable()
//				, 0, observationTable.getColumns().size());
//		ObservationRow<Double> row = observationTable.getUpperTableRow(contextWord.getEmptyWord());
//		// in case of f(epsilon) = 0
//		if (row.getValues().get(0).equals(0.0)){
//			System.out.println("Is target function equal to zero: 1/0?");
//			boolean answer = InputHelper.getInputAnswer();
//			if(! answer) {
//				System.out.println("Please input a nonzero counterexample :");
//			}
//			Word word = InputHelper.getCeWord(contextWord);
//			observationTable.addColumn(word);
//			//f(epsilon, word)
//			Query<Double> query = new QuerySimple<>(row,word,1);
//			membershipOracle.answerMembershipQuery(query);
//			row.set(1, query.getQueryAnswer());
//			observationTable.addUpperRow(word);
//			//f(word, epsilon)
//			row = observationTable.getUpperTableRow(word);
//			row.set(0, query.getQueryAnswer());
//			//f(word, word)
//			query = new QuerySimple<>(row,word,1);
//			membershipOracle.answerMembershipQuery(query);
//			row.set(1, query.getQueryAnswer());
//			//add new lower rows
//			for(int letterNr = 0; letterNr < contextWord.getNumLetters(); letterNr ++) {
//				Word newWord = word.append(letterNr);
//				observationTable.addLowerRow(newWord);}
//			}
//		// ask initial queries for lower table
//			size = observationTable.getsize();
//			processMembershipQueries(observationTable.getLowerTable()
//				, 0, observationTable.getColumns().size());

		
	}
	
	private void processMembershipQueries(List<ObservationRow> rows,
			int colOffset, int length){
		List<Query<Double>> membershipQueries = new ArrayList<>();
//		List<Word> columns = observationTable.getColumns();
//		int endNr = length + colOffset;
//		for(ObservationRow<Double> row : rows){
//			for (int colNr = colOffset; colNr < endNr; colNr ++)
//				membershipQueries.add(new QuerySimple<>(row, columns.get(colNr), colNr));
//		}
		processMembershipQueries(membershipQueries);
	}
	
	private void processMembershipQueries(List<Query<Double>> queries){
		membershipOracle.answerMembershipQueries(queries);
		for(Query<Double> query : queries){
			Double result = query.getQueryAnswer();
			assert result != null;
//			ObservationRow<Double> row = query.getQueriedRow();
//			row.set(query.getQueriedColumnIndex(), result);
		}
	}
	

	
	@Override
	public Automaton getHypothesis() {
		//setting Fx, for x in upper table
		A =new Matrix (size, size);
		matrixes = new ArrayList<>();
		mat = new ArrayList<>();
		int rowNr = 0;
 		for (ObservationRow row: observationTable.getUpperTable()){
 			for(int colNr = 0; colNr < size; colNr ++) {
// 				A.set(rowNr, colNr, row.getValues().get(colNr));
 			}
 			rowNr ++;
 		}
		for(int letterNr = 0; letterNr < contextWord.getNumLetters(); letterNr ++) {
			rowNr = 0;
			//Corresponding Fx·a setting B
			Matrix B = new Matrix (size, size); 		
			// u(a), for a in alphabet
			Matrix solution = new Matrix(size, size);
			for (ObservationRow row: observationTable.getUpperTable()){
				Word rowname = row.getWord().append(letterNr);
				row = observationTable.getLowerTableRow(rowname);
				for(int colNr = 0; colNr < size; colNr ++) {
//	 				B.set(rowNr, colNr, row.getValues().get(colNr));
	 			}
				rowNr ++;
			}
			
			//Solve X*A = B
			System.out.println("The matrix of letter " + contextWord.getLetterWord(letterNr).toStringWithAlphabet()
					+ " is:");
			solution = A.solveTranspose(B).transpose();
			solution.print(4, 2);
			mat.add(solution);
			matrixes.add(B);
		}
		gama = new Matrix(1, size, 0);
		MultiplicityAutomaton ma = new MultiplicityAutomaton(contextWord);
		ma.setStatesWeight(A, size);
		for (int state = 0; state < size; state ++){
			ma.addNewState(state);
			for(int letterNr = 0; letterNr < contextWord.getNumLetters(); letterNr ++) {
				// matrix u(a)
				Matrix u = mat.get(letterNr);
				for(int toState = 0; toState < size; toState ++){
					ma.addTransition(letterNr, toState, u.get(state, toState));}
				}
			// one initial state
			if (state == 0) ma.setInitial();
			// accepting state x with f(x)!= 0
			if (A.get(state, 0) != 0) {
				ma.setAccepting();
				gama.set(0, state, A.get(state, 0));
			}
			ma.addNewStateEnd();
		}
		System.out.println("Vector gama is: " );
		gama.print(4, 2);
		return ma;
	}
	@Override
	public void refineHypothesis(Query<Double> query) {
		Word ceWord = query.getQueriedWord();
		// get u(w)'s first row for every w's prefix
		List<Matrix> rowvectors = new ArrayList<>();
		Matrix vector = new Matrix(1, size, 0);
		vector.set(0, 0, 1);
		for (int i = 0; i < ceWord.length(); i ++){
			vector = vector.times(mat.get(ceWord.getLetter(i)));
			rowvectors.add(vector.copy());

		}
		Matrix result = rowvectors.get(ceWord.length()-1).times(gama.transpose());
		System.out.println("The word: " +ceWord.toStringWithAlphabet()
		+ " has hypothesis value :  " + 
		Double.parseDouble(new DecimalFormat("#.####").format(result.get(0, 0))));
		System.out.println("Is it a counterexample? 1/0");
		boolean answer = InputHelper.getInputAnswer();
		if (!answer) return;
		int index = binarySearch(0, ceWord.length()-1,
				ceWord, rowvectors);
		if (index == ceWord.length()-1){
			result = rowvectors.get(ceWord.length()-1).times(gama.transpose());
			System.out.println("The word: " +ceWord.toStringWithAlphabet()
			+ " is not a counterexample,") ;
			System.out.println("whose hypothesis value is:  " + 
					result.get(0, 0));
		}
		else{
		Word addWord = ceWord.getPrefix(index+1);
		observationTable.addUpperRow(addWord);
		List<Query<Double>> membershipQueries = new ArrayList<>();
		for (int colNr = 0; colNr < size; colNr ++)
//			membershipQueries.add(new QuerySimple<>(observationTable.getUpperTableRow(addWord),
//					observationTable.getColumns().get(colNr), colNr));
		for(int letterNr = 0; letterNr < contextWord.getNumLetters(); letterNr ++) {
			Word newWord = ceWord.getPrefix(index+1).append(letterNr);
			observationTable.addLowerRow(newWord);
//			for (int colNr = 0; colNr < size; colNr ++)
//				membershipQueries.add(new QuerySimple<>(observationTable.getLowerTableRow(newWord),
//						observationTable.getColumns().get(colNr), colNr));
		}
		processMembershipQueries(membershipQueries);
//		observationTable.addColumn(contextWord.getLetterWord(ceWord.getLetter(index+1))
//				.concat(observationTable.getColumns().get(colIndex)));
		//new size after adding new row
		size = observationTable.getsize();
		processMembershipQueries(observationTable.getUpperTable(), size-1, 1);
		processMembershipQueries(observationTable.getLowerTable(), size-1, 1);}
	}
		
	public String toString() {
		return observationTable.toString();
	}
	
	public int binarySearch (int lindex, int rindex,
			Word word, List<Matrix> vectors){
		for(int i = 0; i< 2; i++)
			mat.get(i).print(4, 2);
		int	midindex=(lindex + rindex)/2;
		boolean check1stCond,check2ndCond;
		//check the 1st condition
		ObservationRowMult row = 
				new ObservationRowMult(word.getPrefix(midindex+1));
		//restore values for F(word(midindex)) and later F(word(midindex+1))	
		Matrix vector = new Matrix(1, size);
		Matrix sum = new Matrix(1, size, 0);
		for (int i = 0; i < size; i ++){
//			Query<Double> query = new QuerySimple<>(row,
//					observationTable.getColumns().get(i),i);
//			vector.set(0, i, membershipOracle.answerMembershipQuery(query));
			sum.plusEquals(A.getMatrix(i, i, 0, size-1)
					.times(vectors.get(midindex).get(0, i))) ;
		}
		check1stCond = compare(vector, sum);
		if(check1stCond){
			if(lindex == rindex)
				return lindex;
			
			else{
			System.out.println("condition 1");
			//check the 2nd condition
			sum = new Matrix (1, size, 0);
			//getprifix(midindex+2): return word[0, midindex+1]
			row = new ObservationRowMult(word.getPrefix(midindex+ 2));
			//matrix consisting of Fxa, getletter(midindex+1) corresponding to a
			Matrix Fxa = matrixes.get(word.getLetter(midindex+1));
			Fxa.print(4, 2);
			for (int i = 0; i < size; i ++){
//				Query<Double> query = new QuerySimple<>(row,
//						observationTable.getColumns().get(i),i);
//				vector.set(0, i, membershipOracle.answerMembershipQuery(query));
				sum.plusEquals(Fxa.getMatrix(i, i, 0, size-1)
						.times(vectors.get(midindex).get(0, i))) ;}	
			vector.print(4, 2);
			sum.print(4, 2);
			check2ndCond = !compare(vector, sum);}
			if(check2ndCond){
				System.out.println("condition 2 : midindex:" + midindex);
				//find the column index
				for (int i = 0; i < size; i++)
					if (sum.get(0, i) != vector.get(0, i)){
						setIndex(i);
						break;}
				return midindex;}
			else
				if (rindex == lindex + 1) 	
					return rindex;
				else
					return binarySearch(midindex, rindex,  word, vectors);}
		else
			//rindex == lindex + 1, doesn't satisfy condition 1,
			if(lindex == midindex)
				return lindex;
			return binarySearch(lindex, midindex,  word, vectors);
	}
	
	public void setIndex(int i){
		this.colIndex = i;
	}
	
	public boolean compare(Matrix A, Matrix B){
		if (A.getRowDimension() == B.getRowDimension() && A.getColumnDimension() == B.getColumnDimension()){
			for (int i = 0; i < A.getRowDimension(); i ++)
				for (int j = 0; j < A.getColumnDimension(); j ++)
					if ( Double.parseDouble(
							new DecimalFormat("#.####").format(A.get(i, j)))
							!= Double.parseDouble(
							new DecimalFormat("#.####").format(B.get(i, j))))
						return false;
			return true;}
		return false;
	}
	
	public Matrix mod2(Matrix A){
		Matrix B = A.copy();
		for (int i = 0; i < A.getRowDimension(); i ++)
			for(int j = 0; j < A.getColumnDimension(); j ++)
				B.set(i, j, ((int)A.get(i, j)) % 2);
		return B;
	}
		
	public void printh(int length){
		for (int i = 0; i < Math.pow(2, length); i ++){
			int temp = i;
			Word ceWord = contextWord.getEmptyWord();
			int num = (int) Math.pow(2, length);
			for(int j = 0; j < length; j ++){
				if (temp >= (num /= 2)){
					ceWord = ceWord.append(1);
					temp -= num;}
				else
					ceWord = ceWord.append(0);
		}
	Matrix vector = new Matrix(1, size, 0);
	vector.set(0, 0, 1);
	for (int j = 0; j < ceWord.length(); j ++){
		vector = vector.times(mat.get(ceWord.getLetter(j)));
	}
	Matrix result = vector.times(gama.transpose());
	System.out.println("The word: " +ceWord.toStringWithAlphabet()
	+ " has hypothesis value :  " + 
	Double.parseDouble(new DecimalFormat("#.####").format(result.get(0, 0))));
	}
		}
}

