/* Written by Yong Li, Depeng Liu                                       */
/* Copyright (c) 2016                  	                               */
/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
 /* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License    */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.*/

package cn.ac.ios.learner.table;

import java.io.ByteArrayOutputStream;
import java.util.List;

import cn.ac.ios.automata.words.ContextWord;
import cn.ac.ios.automata.words.Word;
import cn.ac.ios.table.ExprValue;
import cn.ac.ios.table.ExprValueWord;
import cn.ac.ios.table.ExprValueWordPair;
import cn.ac.ios.table.HashableValue;
import cn.ac.ios.table.HashableValueBoolean;
import cn.ac.ios.table.HashableValueBooleanPair;
import cn.ac.ios.table.ObservationRow;
import cn.ac.ios.table.ObservationTableAbstract;
import cn.ac.ios.table.ObservationTablePrinterBoolean;

public abstract class ObservationTableDFA extends ObservationTableAbstract {
	// empty table
	protected ObservationTableDFA(ContextWord contextWord) {
		super(contextWord);
	}
	
	
	@Override
	public ObservationRowDFA getUnclosedLowerRow() {
		for(ObservationRow lowerRow : lowerTable) {
			boolean found = false;
			// found equal upper row
			for(ObservationRow upperRow : upperTable) {
				if(lowerRow.valuesEqual(upperRow)) { 
					found = true;
					break;
				}
			}
			if(!found) {
				return (ObservationRowDFA)lowerRow;
			}
		}
		return null;
	}

	// row(s1) = row(s2) then it should be row(s1.a) = row(s2.a) for every a
	@Override
	public Word getInconsistentColumn() {
		return null;
	}
 
	public String toString() {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        try {
        	ObservationTablePrinterBoolean.print(this, out);
            return out.toString();
        } catch (Exception e) {
            return "ERROR";
        }
	}
	
    // simple add methods, do not contain any check 
	@Override
	public ObservationRowDFA addLowerRow(Word word) {
		ObservationRowDFA row = getRowInstance(word);
		lowerTable.add(row);
		return row;
	}

	@Override
	public ObservationRowDFA addUpperRow(Word word) {
		ObservationRowDFA row = getRowInstance(word);
		upperTable.add(row);
		return row;
	}
	
	public abstract ObservationRowDFA getRowInstance(Word word);

}
