/* Written by Yong Li, Depeng Liu                                       */
/* Copyright (c) 2016                  	                               */
/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
 /* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License    */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.*/

package cn.ac.ios.roll;

import automata.FiniteAutomaton;
import cn.ac.ios.automata.Automaton;
import cn.ac.ios.automata.words.Alphabet;
import cn.ac.ios.automata.words.ContextWord;
import cn.ac.ios.learner.Learner;
import cn.ac.ios.learner.fdfa.LearnerFDFA;
import cn.ac.ios.learner.fdfa.table.LearnerFDFATablePeriodic;
import cn.ac.ios.learner.fdfa.table.LearnerFDFATableRecurrent;
import cn.ac.ios.learner.fdfa.table.LearnerFDFATableSyntactic;
import cn.ac.ios.learner.fdfa.tree.LearnerFDFATreePeriodic;
import cn.ac.ios.learner.fdfa.tree.LearnerFDFATreeRecurrent;
import cn.ac.ios.learner.fdfa.tree.LearnerFDFATreeSyntactic;
import cn.ac.ios.learner.lstar.LearnerLStar;
import cn.ac.ios.learner.omega.buechi.LearnerOmegaBuechi;
import cn.ac.ios.learner.table.dfa.LearnerTableDFA;
import cn.ac.ios.learner.tree.dfa.LearnerTreeDFA;
import cn.ac.ios.options.Options;
import cn.ac.ios.oracle.AutomatonGenerator;
import cn.ac.ios.oracle.EquivalenceOracleImpl;
import cn.ac.ios.oracle.MembershipOracleImpl;
import cn.ac.ios.oracle.Minimizer;
import cn.ac.ios.oracle.Teacher;
import cn.ac.ios.oracle.TeacherDK;
import cn.ac.ios.oracle.TeacherRabit;
import cn.ac.ios.oracle.TeacherSample;
import cn.ac.ios.query.EquivalenceOracle;
import cn.ac.ios.query.MembershipOracle;
import cn.ac.ios.query.Query;
import cn.ac.ios.roll.sampler.FASampler;
import cn.ac.ios.translator.Translator;
import cn.ac.ios.translator.TranslatorBuechi;
import cn.ac.ios.translator.TranslatorFDFAOver;
import cn.ac.ios.translator.TranslatorFDFAUnder;
import cn.ac.ios.translator.TranslatorSimple;
import cn.ac.ios.util.AutomatonPrinter;
import cn.ac.ios.util.BuechiBuilder;
import cn.ac.ios.util.HanoiParser;
import cn.ac.ios.util.InputHelper;
import cn.ac.ios.util.Timer;
import cn.ac.ios.util.UtilAutomaton;
import cn.ac.ios.value.ContextValue;
import cn.ac.ios.value.Type;
import cn.ac.ios.value.Value;
import oracle.EqResult;


/**
 * This is the main class for Jalf, we have two mode
 * 
 * 1. interactive mode, you can teach Jalf by yourself
 * 2. automatic automata learning from teacher Rabit tool
 * http://www.languageinclusion.org/doku.php?id=tools
 *
 * TODO add automatic learning for regular language
 */
public class ROLL 
{
	private static boolean isOmegaLearning = false;
	private static double ep;
	private static double del;
	private static int numST;
	private static int numTests;
	private static String outputFile = null;
	
    public static void main(String[] args)
    {
    	// parse input arguments
    	for(int i = 0; i < args.length; i ++) {
    		
    		if(args[i].compareTo("-h")==0) {
    			helper();
    		}
    	}
    	
    	if(args.length < 2) {
    		helper();
    	}

    	String autfile = null;
    	// for learning;
    	for(int i = 0; i < args.length; i ++) {
    		if(args[i].compareTo("-aut") == 0) {
    			Options.modeInteractive = false;
    			Options.modeAutomatic = true;
    			Options.teacherSampling = false;
    			continue;
    		}
    		if(args[i].compareTo("-int") == 0) {
    			Options.modeInteractive = true;
    			Options.modeAutomatic = false;
    			Options.teacherSampling = false;
    			continue;
    		}
			
			if(args[i].compareTo("-sameq") == 0) {
				Options.teacherSampling = true;
    			Options.modeInteractive = false;
    			Options.modeAutomatic = true;
    			ep = Double.parseDouble(args[i+1]);
    			del = Double.parseDouble(args[i+2]);
    			i += 2;
    			continue;
			}
			
			if(args[i].compareTo("-v")==0){
				Options.verbose=true;
				continue;
			}
			
			if(args[i].compareTo("-out")==0){
				outputFile = args[i+1];
				i += 1;
				continue;
			}
			
			if(args[i].compareTo("-table") == 0) {
    			Options.learnerTree = false;
    			Options.learnerTable = true;
    			continue;
			}
			if(args[i].compareTo("-tree") == 0) {
    			Options.learnerTree = true;
    			Options.learnerTable = false;
    			continue;
			}
			if(args[i].compareTo("-lstar") == 0) {
				Options.learnerLStar = true;
				Options.learnerMultiplicity = false;
				Options.learnerDFA = false;
				Options.learnerNFA = false;
				Options.learnerBuechi = false;
				Options.learnerPeriodic = false;
				Options.learnerRecurrent = false;
				Options.learnerSyntactic = false;
				Options.lazyEqCheck = false;
				continue;
			}
			if(args[i].compareTo("-dfa") == 0) {
				Options.learnerDFA = true;
				Options.learnerLStar = false;
				Options.learnerMultiplicity = false;
				Options.learnerNFA = false;
				Options.learnerBuechi = false;
				Options.learnerPeriodic = false;
				Options.learnerRecurrent = false;
				Options.learnerSyntactic = false;
				Options.lazyEqCheck = false;
				continue;
			}
			if(args[i].compareTo("-ldollar") == 0) {
				Options.learnerLStar = false;
				Options.learnerMultiplicity = false;
				Options.learnerNFA = false;
				Options.learnerDFA = false;
				Options.learnerBuechi = true;
				Options.learnerPeriodic = false;
				Options.learnerRecurrent = false;
				Options.learnerSyntactic = false;
				isOmegaLearning = true;
				continue;
			}
			if(args[i].compareTo("-syntactic") == 0) {
				Options.learnerLStar = false;
				Options.learnerMultiplicity = false;
				Options.learnerNFA = false;
				Options.learnerDFA = false;
				Options.learnerBuechi = false;
				Options.learnerPeriodic = false;
				Options.learnerRecurrent = false;
				Options.learnerSyntactic = true;
				isOmegaLearning = true;
				continue;
			}
			if(args[i].compareTo("-recurrent") == 0) {
				Options.learnerLStar = false;
				Options.learnerMultiplicity = false;
				Options.learnerDFA = false;
				Options.learnerNFA = false;
				Options.learnerBuechi = false;
				Options.learnerPeriodic = false;
				Options.learnerRecurrent = true;
				Options.learnerSyntactic = false;
				isOmegaLearning = true;
				continue;
			}
			if(args[i].compareTo("-periodic") == 0) {
				Options.learnerLStar = false;
				Options.learnerMultiplicity = false;
				Options.learnerDFA = false;
				Options.learnerNFA = false;
				Options.learnerBuechi = false;
				Options.learnerPeriodic = true;
				Options.learnerRecurrent = false;
				Options.learnerSyntactic = false;
				isOmegaLearning = true;
				continue;
			}
			
			if(args[i].compareTo("-under") == 0) {
				Options.buechiBuildLower = true;
				continue;
			}
			if(args[i].compareTo("-sim") == 0) {
				Options.simulation = true;
				continue;
			}
			
			if(args[i].compareTo("-min") == 0) {
				Options.minimizationBuechi = true;
				continue;
			}
			
			if(args[i].compareTo("-over") == 0) {
				Options.buechiBuildLower = false;
				Options.outLDBA = false;
				continue;
			}

			
			if(args[i].compareTo("-lazyeq") == 0) {
				Options.lazyEqCheck = true;
				continue;
			}

			
			if(args[i].compareTo("-samfdfa") == 0) {
				Options.sampleFdfa = true;
				continue;
			}
			
			if(args[i].compareTo("-ldba") == 0) {
				Options.outLDBA = true;
				continue;
			}
			
			if(args[i].endsWith(".ba")) {
				autfile = args[i];
				continue;
			}
			
			if(args[i].endsWith(".hoa")) {
				autfile = args[i];
				continue;
			}
			
			if(args[i].compareTo("-t") == 0) {
				Options.test = true;
				numTests = Integer.parseInt(args[i+1]);
				numST = Integer.parseInt(args[i + 2]);
				i += 2;
				continue;
			}
    	}
    	
    	// check whether we need to output ldba
    	if(Options.outLDBA) {
    		if(! Options.buechiBuildLower) {
    			System.out.println("Buechi construction is not under construction. No LDBA output.");
    			Options.outLDBA = false;
    		}
    	}
    	
    	Statistics.reset();
    	
    	Learner<Automaton, Boolean> learner = null;
		MembershipOracle<Boolean> membershipOracle = null;
		if(Options.test) { // do 20 tests?
			for(int n = 0; n < numTests; n ++) {
				FiniteAutomaton aut = AutomatonGenerator.getRandomAutomaton(numST, 3);
				try{
					learnInputAutomaton(aut);
				}catch (Exception e)
				{
					e.printStackTrace();
					aut.saveAutomaton("bug.ba");
					System.exit(-1);
				}
				
				System.out.println("Done for case " + (n + 1));
			}
		}else if(Options.modeInteractive) {
			ContextWord contextWord = initializeAlphabet();
    		membershipOracle = new MembershipOracleImpl();
    		learner = pickOneLearner(contextWord, membershipOracle);
    		if(learner == null) {
    			System.err.println("Error, must specify one learner.");
    			System.exit(-1);
    		}
    		doInteractiveMode(contextWord, learner);
    	}else {
    		// performance evaluation process
    		FiniteAutomaton targetAut = loadAutomaton(autfile);
    		Statistics.targetAut = targetAut;
    		Timer timer = new Timer();
    		timer.start();
    		try{
    			learnInputAutomaton(targetAut);
    		}catch(Exception e) {
    			e.printStackTrace();
    		}finally {
    			AutomatonPrinter.print(targetAut, System.out);
    		}
    		timer.stop();
    		Statistics.timeTotal = timer.getTimeElapsed();
    		Statistics.print();
    	}
    	
    }
    
    private static FiniteAutomaton loadAutomaton(String autfile) {
    	if(autfile.endsWith(".ba")) {
    		return new FiniteAutomaton(autfile);
    	}else {
    		HanoiParser parser = new HanoiParser(autfile);
    		return parser.getAutomaton();
    	}
    }
    
    private static ContextWord initializeAlphabet() {
		ContextValue contextValue = new ContextValue();
		ContextWord contextWord = new ContextWord(contextValue);
		System.out.println("Please input the number of letters: ");
		int numLetters = InputHelper.getInterger();
		Type typeObject = contextValue.newTypeObject(String.class);
		contextValue.newTypeLetter(String.class);
		Value valueLetter = typeObject.newValue();
		Alphabet alphabet = contextWord.getAlphabet();
		for(int letterNr = 0; letterNr < numLetters; letterNr ++) {
			System.out.print("Please input the " + (letterNr + 1) + "th letter: ");
			String letter = InputHelper.getLetter();
			valueLetter.set(letter);
			alphabet.add(valueLetter.clone());
		}
		alphabet.setImmutable();
		valueLetter.set("");
		contextWord.setLetterSplitter(valueLetter);
		return contextWord;
    }
    
    private static void learnInputAutomaton(FiniteAutomaton aut) {
		ContextValue contextValue = new ContextValue();
		ContextWord contextWord = new ContextWord(contextValue);
		Teacher<FiniteAutomaton, EqResult, Boolean> teacher = null;
		
		if(isOmegaLearning) {
			if(Options.teacherSampling) teacher = new TeacherSample(contextWord, aut, ep, del);
			else teacher = new TeacherRabit(contextWord, aut);
		}else {
			teacher = new TeacherDK(contextWord, aut);
		}
		
		MembershipOracle<Boolean> membershipOracle = teacher;
		Learner<Automaton, Boolean> learner = pickOneLearner(contextWord, membershipOracle);
		doAutomaticMode(teacher, learner);
    }

	private static Learner<Automaton, Boolean> pickOneLearner(ContextWord contextWord,
    		MembershipOracle<Boolean> membershipOracle) {
    	
    	if(Options.learnerTable) {
    		if(Options.learnerLStar) return new LearnerLStar(contextWord, membershipOracle);
    		if(Options.learnerDFA) return new LearnerTableDFA(contextWord, membershipOracle);
    		if(Options.learnerPeriodic) return new LearnerFDFATablePeriodic(contextWord, membershipOracle);
    		if(Options.learnerRecurrent) return new LearnerFDFATableRecurrent(contextWord, membershipOracle);
    		if(Options.learnerSyntactic) return new LearnerFDFATableSyntactic(contextWord, membershipOracle);
    		if(Options.learnerBuechi) return new LearnerOmegaBuechi(contextWord, membershipOracle, false);
    	}else {
    		if(Options.learnerLStar) {
    			System.err.println("LStar does not support Tree-based Algorithm.");
    			System.exit(-1);
    		}
    		if(Options.learnerDFA) return new LearnerTreeDFA(contextWord, membershipOracle);
    		if(Options.learnerPeriodic) return new LearnerFDFATreePeriodic(contextWord, membershipOracle);
    		if(Options.learnerRecurrent) return new LearnerFDFATreeRecurrent(contextWord, membershipOracle);
    		if(Options.learnerSyntactic) return new LearnerFDFATreeSyntactic(contextWord, membershipOracle);
    		if(Options.learnerBuechi) return new LearnerOmegaBuechi(contextWord, membershipOracle, true);
    	}
    	return null;
    }
    
    public static void helper() {
		System.out.println(
				"ROLL v1 (Regular Omega Language Learning)");
		System.out.println(
				"Usage: java -jar ROLL.jar [aut.ba, aut.hoa] [options]");

		System.out.println("Recommended use: java -jar ROLL.jar -int -lstar");
		System.out.println("             or: java -jar ROLL.jar aut.ba -aut -table -periodic -under");
		System.out.println("             or: java -jar ROLL.jar aut.hoa -aut -table -periodic -under");
		System.out.println("             or: java -jar ROLL.jar -t 3 3 -aut -table -periodic -under");
		System.out.println("\nOptions:");
		System.out.println("-h: Show this page.");
		System.out.println("-v: Verbose mode.");
		System.out.println("-out filename: Output learned automaton in filename.");
		System.out.println(
				"-t k n: Test ROLL with k randomly generated Buechi automata of n states.");
		System.out.println(
				"-int: Play ROLL in an interactive way, you play the role as a teacher.");
		System.out.println(
				"-aut: Use RABIT or DK package tool as the teacher. ");
		System.out.println(
				"-sameq e d: Sampling as the teacher to check equivalence of two BA automata.\n"
				+ "             e - the probability that equivalence check is not correct\n"
				+ "             d - the probability of the confidence for equivalence check");
		System.out.println(
				"-tree: Use tree-based data structure in the algorithm (Default).");
		System.out.println(
				"-table: Use table-based data structure in the algorithm.");
		System.out.println(
				"-lstar: Use classic L* algorithm.");
		System.out.println(
				"-dfa: Use column based DFA learning algorithm.");
		System.out.println(
				"-ldollar:  Use CNP algorithm and L$ automata to learn Omega regular language.");
		System.out.println(
				"-periodic: Use peridoc FDFA to learn Omega regular language.");
		System.out.println(
				"-recurrent: Use recurrent FDFA to learn Omega regular language.");
		System.out.println(
				"-syntactic: Use syntactic FDFA to learn Omega regular language.");
		System.out.println(
				"-over: Use over Buechi automaton construction for FDFA. (Default)");
		System.out.println(
				"-under: Use under Buechi automaton construction for FDFA.");
//		System.out.println(
//				"-samfdfa: Sampling the FDFA before equivalence check. (Not yet supported)");
		System.out.println(
				"-sim: Use Similation provided by RABIT to reduce automata before intersection of DFAs.");
		System.out.println(
				"-min: Use Minimization provided by RABIT to reduce Buechi automata before equivalence check.");
		System.out.println(
				"-lazyeq: Equivalence check as the last resort. ");
		System.out.println(
				"-ldba: Output the final Buechi as Limit Deterministic Buechi Automaton (with -under). ");
		System.exit(0);
    }
    
    public static void doInteractiveMode(ContextWord contextWord, Learner<Automaton, Boolean> learner) {
		System.out.println("Start learning...");
		learner.startLearning();
		boolean result = false;
		while(! result ) {
			Automaton model = learner.getHypothesis();
			EquivalenceOracle<Automaton, Boolean> equivalenceOracle = new EquivalenceOracleImpl();
			result = equivalenceOracle.answerEquivalenceQuery(model);
			if(result == true) break;
			Query<Boolean> ceQuery = null;
			if(isOmegaLearning) {
				ceQuery = InputHelper.getOmegaCeWord(contextWord);
			}else {
				ceQuery = InputHelper.getCeWord(contextWord);
			}
			learner.refineHypothesis(ceQuery);
		}
    }
    
    // TODO make full use of counterexample that teacher returned
    private static void doAutomaticMode(Teacher<FiniteAutomaton, EqResult, Boolean> teacher
    		, Learner<Automaton, Boolean> learner) {
		System.out.println("Start learning...");
		Timer timer = new Timer();
		timer.start();
		learner.startLearning();
		timer.stop();
		Statistics.timeLearner += timer.getTimeElapsed();
		boolean result = false;
		while(! result ) {
			if(Options.verbose) System.out.println("learner output: " + learner.toString());
			EquivalenceOracle<FiniteAutomaton, Query<EqResult>> equivalenceOracle = teacher;
			
			Query<EqResult> query = null;
			// make it mouch more precise by sampling before equivalence check
			// if we get a counterexample by sampling, then we do not have to use
			// equivalence check this time
			if(learner instanceof LearnerFDFA && Options.sampleFdfa) {
				timer.start();
				FASampler sampler = new FASampler(learner, teacher);
				sampler.sample();
				query = sampler.getCounterexample();
				timer.stop();
				Statistics.timeSampling += timer.getTimeElapsed();
				++ Statistics.numSamplingTried;
			}
			
			FiniteAutomaton hypothesis = null;
			// do not get counterexample by sampling or we do not use sampling
			if(query == null) {
				hypothesis = UtilAutomaton.convertToRabitAut(learner);
				if(Options.verbose) {
					AutomatonPrinter.print(hypothesis, System.out);
//				    hypothesis.saveAutomaton("testmem.ba");
				}
				if(Options.minimizationBuechi) {
					timer.start();
					// maybe we can set the number of step to look ahead
					hypothesis = Minimizer.minimizeBuechi(hypothesis, 12);
					timer.stop();
					Statistics.timeMinimizationBuechi += timer.getTimeElapsed();
				}
				query = equivalenceOracle.answerEquivalenceQuery(hypothesis);
			}

			// get out of the loop
			if(query.getQueryAnswer().isEqual == true) {
				Statistics.setLearnerFDFA(learner);
				Statistics.hypothesis = hypothesis;
				if(outputFile != null) hypothesis.saveAutomaton(outputFile);
				break;
			}
			
			// lazy equivalence check is implemented here
			Translator translator = getTranslator(learner, teacher);
			translator.setQuery(query);
			while(translator.canRefine()) {
				Query<Boolean> ceQuery = translator.translate();
				timer.start();
				learner.refineHypothesis(ceQuery);
				timer.stop();
				Statistics.timeLearner += timer.getTimeElapsed();
				if(Options.verbose) System.out.println("learner output: " + learner.toString());
				// if do not set lazy eq check or it is learnerBuechi
				if(! Options.lazyEqCheck) break;
			}

		}
		
		if(Options.outLDBA && (learner instanceof LearnerFDFA)) {
			Statistics.ldba = BuechiBuilder.buildLDBA((LearnerFDFA)learner);
			if(Options.verbose) { 
				Query<EqResult> test = teacher.answerEquivalenceQuery(Statistics.ldba);
				System.out.println("LDBA equivalence: " + test.getQueryAnswer().isEqual);
			}
		}
		
		System.out.println("Learning completed...");
    }
    
    private static Translator getTranslator(Learner<Automaton, Boolean> learner
    		, MembershipOracle<Boolean> membershipOracle) {
    	if(learner instanceof LearnerFDFA && Options.buechiBuildLower) {
    		return new TranslatorFDFAUnder(learner);
    	}
    	if(learner instanceof LearnerFDFA && !Options.buechiBuildLower) {
    		return new TranslatorFDFAOver(learner, membershipOracle);
    	}
    	if(learner instanceof LearnerOmegaBuechi) {
    		return new TranslatorBuechi(learner);
    	}
    	return new TranslatorSimple(learner);
    }
}
