/* Written by Yong Li, Depeng Liu                                       */
/* Copyright (c) 2016                  	                               */
/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
 /* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License    */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.*/

package cn.ac.ios.value;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.RoundingMode;

// modified from the IscasMC (http://iscasmc.ios.ac.cn/) by Ernst Moritz Hahn

class ValueFraction implements Value{

    private final static BigInteger BI_ZERO = BigInteger.ZERO;
    private final static BigInteger BI_ONE = BigInteger.ONE;
    private final static BigInteger BI_MONE = BigInteger.ONE.negate();
    private final static BigInteger BI_TEN = BigInteger.TEN;
    
    private BigInteger numerator;
    private BigInteger denominator;
    private Type type;
    private boolean immutable;
    // 1, 0, -1, 10
    ValueFraction(Type type, BigInteger num, BigInteger den) {
    	assert type != null;
        this.type = type;
        assert num != null;
        assert den != null;
        this.numerator = num;
        this.denominator = den;
    }
    
    ValueFraction(Type type, BigInteger num) {
    	this(type, num, BI_ONE);
    }
    
    ValueFraction(Type type, int num) {
    	this(type, BigInteger.valueOf(num), BI_ONE);
    }
    
    ValueFraction(Type type) {
        this(type, BI_ZERO, BI_ZERO);
    }


    @Override
    public Value clone() {
        return new ValueFraction(getType(), numerator, denominator);
    }

    @Override
    public boolean equals(Object obj) {

    }

    @Override
    public int hashCode() {
    	// TODO
        return 0;
    }

    @Override
    public String toString() {
	return null;
    }

    
    @Override
    public boolean isFraction() {
    	return true;
    }
    
    @Override
    public void add(Value operand1, Value operand2)  {

    }
    
    @Override
    public void subtract(Value operand1, Value operand2)
            {
    }
    
    @Override
    public void inverse(Value operand) {

    }

    @Override
    public double getDouble() {
        return toDouble();
    }
    
    @Override
    public void multiply(Value operand1, Value operand2) {
    }
    
    @Override
    public void divide(Value operand1, Value operand2) {
    }
    
    @Override
    public boolean isOne() {
        return false;
    }
    
    @Override
    public boolean isZero() {
        return false;
    }

    
    @Override
    public void set(String value) {

    }
    
    @Override
    public void set(int value) {

    }
    
    @Override
    public void set(Value value) {
    	assert value != null;
    }
 

    @Override
    public Type getType() {
        return type;
    }

    @Override
    public void setImmutable() {
        this.immutable = true;
    }
    
    @Override
    public boolean isImmutable() {
        return immutable;
    }


}
